@@ -18,6 +18,7 @@ except ModuleNotFoundError as e:
     pass
 
 import _damo_subproc
+import _damo_sysinfo
 import _damon
 import damo_pa_layout
 
@@ -314,8 +315,10 @@ def damos_options_to_quotas(quotas, goals):
         return None, 'Wrong --damos_quotas (%s, %s)' % (qargs, e)
     return quotas, None
 
-def damos_options_to_scheme(sz_region, access_rate, age, action,
-        apply_interval, quotas, goals, wmarks, target_nid, filters, dests):
+def damos_options_to_scheme(
+        sz_region, access_rate, age, action,
+        apply_interval, quotas, goals, wmarks, target_nid, filters, dests,
+        max_nr_snapshots):
     if quotas != None:
         quotas, err = damos_options_to_quotas(quotas, goals)
         if err is not None:
@@ -333,13 +336,18 @@ def damos_options_to_scheme(sz_region, access_rate, age, action,
     if err != None:
         return None, err
 
+    if max_nr_snapshots is None:
+        stats = _damon.DamosStats()
+    else:
+        stats = _damon.DamosStats(max_nr_snapshots=max_nr_snapshots)
+
     try:
         return _damon.Damos(
                 access_pattern=_damon.DamosAccessPattern(sz_region,
                     access_rate, _damon.unit_percent, age, _damon.unit_usec),
                 action=action, target_nid=target_nid,
                 dests=dests, apply_interval_us=apply_interval, quotas=quotas,
-                watermarks=wmarks, filters=filters), None
+                watermarks=wmarks, filters=filters, stats=stats), None
     except Exception as e:
         return None, 'Wrong \'--damos_*\' argument (%s)' % e
 
@@ -399,6 +407,9 @@ def verify_set_damos_args_len(args):
         args.damos_nr_filters = [len(args.damos_filter)]
     if sum(args.damos_nr_filters) != len(args.damos_filter):
         return 'wrong --damos_nr_filters'
+
+    if len(args.damos_max_nr_snapshots) > nr_schemes:
+        return 'too much --damos_max_nr_snapshots'
     return None
 
 def fillup_default_damos_args(args):
@@ -413,6 +424,8 @@ def fillup_default_damos_args(args):
             nr_schemes - len(args.damos_apply_interval))
     args.damos_quotas += [None] * (nr_schemes - len(args.damos_quotas))
     args.damos_wmarks += [None] * (nr_schemes - len(args.damos_wmarks))
+    args.damos_max_nr_snapshots += [None] * (
+            nr_schemes - len(args.damos_max_nr_snapshots))
 
 def damos_options_to_schemes(args):
     set_args_damos_quotas(args)
@@ -462,7 +475,7 @@ def damos_options_to_schemes(args):
                 args.damos_access_rate[i], args.damos_age[i],
                 args.damos_action[i], args.damos_apply_interval[i],
                 args.damos_quotas[i], qgoals, args.damos_wmarks[i],
-                target_nid, filters, dests)
+                target_nid, filters, dests, args.damos_max_nr_snapshots[i])
         if err != None:
             return [], err
         schemes.append(scheme)
@@ -500,6 +513,64 @@ def damon_target_for(args, idx, ops):
         return None, 'Wrong \'--target_pid\' argument (%s)' % e
     return target, None
 
+def build_sample_control_ops_attrs(args, idx):
+    '''
+    Returns DamonSampleControl, OpsAttrs, and an error
+    '''
+    if args.exp_ops_use_reports[idx] is not None:
+        use_reports = args.exp_ops_use_reports[idx]
+    else:
+        use_reports = False
+    if args.exp_ops_write_only[idx] is not None:
+        write_only = args.exp_ops_write_only[idx]
+    else:
+        write_only = False
+    if args.exp_ops_cpus[idx] is not None:
+        cpus = args.exp_ops_cpus[idx]
+    else:
+        cpus = 'all'
+    if args.exp_ops_tids[idx] is not None:
+        tids = args.exp_ops_tids[idx]
+    else:
+        tids = ''
+    use_sample_control, err = _damo_sysinfo.damon_sysfs_feature_available(
+            'damon_sample_control')
+    if err is not None:
+        return None, None, \
+                'damon_sample_control feature check fail (%s)' % support
+    if use_sample_control is False:
+        try:
+            ops_attrs = _damon.OpsAttrs(
+                    use_reports=use_reports, write_only=write_only, cpus=cpus,
+                    tids=tids)
+        except Exception as e:
+            return None, None, 'invalid ops_attrs arguments (%s)' % e
+        return None, ops_attrs, None
+
+    if use_reports:
+        page_table, page_fault = False, True
+    else:
+        page_table, page_fault = True, False
+    primitives_enabled = _damon.DamonPrimitivesEnabled(
+            page_table=page_table, page_fault=page_fault)
+    sample_filters = []
+    if write_only:
+        sample_filters.append(_damon.DamonSampleFilter(
+            filter_type=_damon.damon_filter_type_write, matching=True,
+            allow=True))
+    if cpus != 'all':
+        sample_filters.append(_damon.DamonSampleFilter(
+            filter_type=_damon.damon_filter_type_cpumask, matching=True,
+            allow=True, cpumask=cpus))
+    if tids != '':
+        sample_filters.append(_damon.DamonSampleFilter(
+            filter_type=_damon.damon_filter_type_threads, matching=True,
+            allow=True, tid_arr=tids))
+    sample_control = _damon.DamonSampleControl(
+            primitives_enabled=primitives_enabled,
+            sample_filters=sample_filters)
+    return sample_control, None, None
+
 def damon_ctx_for(args, idx):
     if args.ops[idx] is None:
         if args.target_pid[idx] is None:
@@ -521,32 +592,14 @@ def damon_ctx_for(args, idx):
     except Exception as e:
         return None, 'invalid nr_regions arguments (%s)' % e
     ops = args.ops[idx]
-    if args.exp_ops_use_reports[idx] is not None:
-        use_reports = args.exp_ops_use_reports[idx]
-    else:
-        use_reports = False
-    if args.exp_ops_write_only[idx] is not None:
-        write_only = args.exp_ops_write_only[idx]
-    else:
-        write_only = False
-    if args.exp_ops_cpus[idx] is not None:
-        cpus = args.exp_ops_cpus[idx]
-    else:
-        cpus = 'all'
-    if args.exp_ops_tids[idx] is not None:
-        tids = args.exp_ops_tids[idx]
-    else:
-        tids = ''
-    try:
-        ops_attrs = _damon.OpsAttrs(
-                use_reports=use_reports, write_only=write_only, cpus=cpus,
-                tids=tids)
-    except Exception as e:
-        return None, 'invalid ops_attrs arguments (%s)' % e
+    sample_control, ops_attrs, err = build_sample_control_ops_attrs(args, idx)
+    if err is not None:
+        return None, 'exp_ops_* handling fail (%s)' % err
 
     try:
-        ctx = _damon.DamonCtx(ops, None, intervals, nr_regions, schemes=[],
-                              ops_attrs=ops_attrs)
+        ctx = _damon.DamonCtx(
+                ops, None, intervals, nr_regions, schemes=[],
+                sample_control=sample_control, ops_attrs=ops_attrs)
         return ctx, None
     except Exception as e:
         return None, 'Creating context from arguments failed (%s)' % e
@@ -855,12 +908,49 @@ def stage_kdamonds(args):
         return None, 'cannot apply kdamonds from args (%s)' % err
     return kdamonds, None
 
+def damos_filter_invalidity(filter):
+    # todo: support non-memcg filters
+    if filter.filter_type != 'memcg':
+        return None
+    memcg_path = filter.memcg_path
+    if not memcg_path.startswith('/'):
+        return 'path should start with "/"'
+    cgroup_mount_path = None
+    with open('/proc/mounts', 'r') as f:
+        for line in f:
+            fields = line.split()
+            if not fields[2] in ['cgroup', 'cgroup2']:
+                continue
+            cgroup_mount_path = fields[1]
+            break
+    if cgroup_mount_path is None:
+        return 'cgroup is not mounted'
+    memcg_realpath = os.path.join(cgroup_mount_path, memcg_path[1:])
+    if not os.path.isdir(memcg_realpath):
+        return '%s not found' % memcg_realpath
+    return None
+
+def infer_start_or_commit_fail_reason(kdamonds):
+    for kidx, kd in enumerate(kdamonds):
+        for cidx, ctx in enumerate(kd.contexts):
+            for sidx, scheme in enumerate(ctx.schemes):
+                for fidx, filter in enumerate(scheme.filters):
+                    reason = damos_filter_invalidity(filter)
+                    if reason is not None:
+                        return ''.join([
+                            'wrong %d/%d/%d/%d-th ' % (kidx, cidx, sidx, fidx),
+                            'kdamond/context/scheme/filter (%s)' % reason])
+    return None
+
 def commit_kdamonds(args, commit_quota_goals_only):
     kdamonds, err = kdamonds_for(args)
     if err:
         return None, 'cannot create kdamonds to commit from args (%s)' % err
     err = _damon.commit(kdamonds, commit_quota_goals_only)
     if err:
+        inferred_reason = infer_start_or_commit_fail_reason(kdamonds)
+        if inferred_reason is not None:
+            err = '%s (%s)' % (err, inferred_reason)
         return None, 'cannot commit kdamonds (%s)' % err
     return kdamonds, None
 
@@ -868,8 +958,14 @@ def turn_damon_on(args):
     kdamonds, err = stage_kdamonds(args)
     if err:
         return err, None
-    return _damon.turn_damon_on(
-            ['%s' % kidx for kidx, k in enumerate(kdamonds)]), kdamonds
+    err = _damon.turn_damon_on(
+            ['%s' % kidx for kidx, k in enumerate(kdamonds)])
+    if err is not None:
+        inferred_reason = infer_start_or_commit_fail_reason(kdamonds)
+        if inferred_reason is not None:
+            err = '%s (%s)' % (err, inferred_reason)
+        return err, None
+    return None, kdamonds
 
 # Commandline options setup helpers
 
@@ -1058,6 +1154,9 @@ def set_damos_argparser(parser, hide_help):
                 '<low mark (permil)>'),
             help='damos watermarks'
             if not hide_help else argparse.SUPPRESS)
+    parser.add_argument(
+            '--damos_max_nr_snapshots', action='append', default=[],
+            help='damos max_nr_snapshots')
     parser.add_argument('--nr_schemes', metavar='<number>', nargs='+', type=int,
                         help='number of schemes for each context (in order)'
                         if not hide_help else argparse.SUPPRESS)
